import warnings
from io import StringIO
from pathlib import Path
from typing import Callable, Iterable, Optional

import numpy as np
import pandas as pd
from aislib.misc_utils import get_logger
from eir.setup.input_setup_modules.setup_omics import read_bim
from joblib import Parallel, delayed
from statsmodels.base.model import LikelihoodModelResults
from statsmodels.formula import api as smf
from statsmodels.tools.sm_exceptions import ConvergenceWarning

logger = get_logger(name=__name__)

warnings.filterwarnings(action="ignore", category=ConvergenceWarning)
warnings.filterwarnings(action="ignore", category=RuntimeWarning)


def get_allele_effects(
    df_genotype: pd.DataFrame,
    df_target: pd.DataFrame,
    bim_file: Path,
    target_type: str,
) -> pd.DataFrame:
    target_name = df_target.columns[0]
    df_combined = pd.concat(objs=[df_target, df_genotype], axis=1)

    df_bim = read_bim(bim_file_path=str(bim_file))
    rs_ids = df_genotype.columns
    allele_maps = get_snp_allele_maps(df_bim=df_bim, snp_ids=rs_ids)

    df_results = get_all_linear_results(
        df=df_combined,
        target_name=target_name,
        allele_maps=allele_maps,
        target_type=target_type,
    )

    if "P>|t|" in df_results.columns:
        df_results = df_results.rename(columns={"P>|t|": "p_value"})
    elif "P>|z|" in df_results.columns:
        df_results = df_results.rename(columns={"P>|z|": "p_value"})
    else:
        raise ValueError("Could not find p value column.")

    return df_results


def get_all_linear_results(
    df: pd.DataFrame,
    target_name: str,
    allele_maps: dict[str, dict[str, str]],
    target_type: str,
) -> pd.DataFrame:
    logger.info("Gathering all linear allele effect results for '%s'.", target_name)

    parallel_worker = Parallel(n_jobs=-1)
    all_results = parallel_worker(
        delayed(_compute_single_snp_effect_wrapper)(
            df=df,
            target_name=target_name,
            allele_maps=allele_maps,
            snp=snp,
            model=target_type,
        )
        for snp in df.columns
        if snp != target_name
    )
    all_results = [i for i in all_results if i is not None]

    if len(all_results) == 0:
        return pd.DataFrame()

    return pd.concat(all_results)


def _compute_single_snp_effect_wrapper(
    df: pd.DataFrame,
    target_name: str,
    snp: str,
    allele_maps: dict[str, dict[str, str]],
    model: str,
) -> Optional[pd.DataFrame]:
    formula = build_basic_fit_formula(target=target_name, snp=snp)

    df_cur = df[[target_name, snp]]
    df_cur_no_na = df_cur[df_cur[snp] != -1]

    fit_func = get_statsmodels_fit_function(target_type=model)

    try:
        result = fit_func(formula=formula, data=df_cur_no_na).fit(disp=0)
        if hasattr(result, "mle_retvals"):
            if not result.mle_retvals["converged"]:
                raise RuntimeError(f"Model did not converge on '{snp}'.")
        elif not np.all(np.isfinite(result.normalized_cov_params)):
            raise RuntimeError(f"Model did not converge on '{snp}'.")

        result_summary = result.summary2().tables[1]
        if not np.all(np.isfinite(result_summary)):
            raise RuntimeError(f"Model resulted in non-finite values on '{snp}'.")

    except Exception as exception:
        logger.info(
            "Failed OLS on '%s' due to exception '%s'. Skipping.",
            snp,
            exception,
        )
        return

    df_result = build_df_from_basic_results(
        results=result,
        allele_maps=allele_maps,
        snp=snp,
    )

    return df_result


def get_statsmodels_fit_function(target_type: str) -> Callable:
    if target_type == "regression":
        func = smf.ols
    elif target_type == "classification":
        func = smf.logit
    else:
        raise ValueError()

    return func


def build_basic_fit_formula(target: str, snp: str) -> str:
    return f"Q('{target}') ~ C(Q('{snp}'))"


def get_snp_allele_maps(
    df_bim: pd.DataFrame, snp_ids: Iterable[str]
) -> dict[str, dict[str, str]]:
    snp_allele_maps = {}
    for rs_id in snp_ids:
        snp_allele_maps[rs_id] = get_snp_allele_nucleotide_map(
            df_bim=df_bim, rs_id=rs_id
        )

    return snp_allele_maps


def get_snp_allele_nucleotide_map(df_bim: pd.DataFrame, rs_id: str) -> dict[str, str]:
    nucleotide_map = {}

    ref = df_bim[df_bim["VAR_ID"] == rs_id]["REF"].item()
    alt = df_bim[df_bim["VAR_ID"] == rs_id]["ALT"].item()

    nucleotide_map["REF"] = ref * 2
    nucleotide_map["HET"] = f"{ref}{alt}"
    nucleotide_map["ALT"] = alt * 2

    return nucleotide_map


def build_df_from_basic_results(
    results: LikelihoodModelResults,
    allele_maps: dict[str, dict[str, str]],
    snp: str,
) -> pd.DataFrame:
    results_as_html = results.summary().tables[1].as_html()
    html_buffer = StringIO(results_as_html)
    df_linear = pd.read_html(io=html_buffer, header=0, index_col=0)[0]
    df_linear.index.name = "allele"

    df_linear_total = compute_total_effect(df=df_linear)

    df_linear_renamed = _rename_linear_regression_index(
        df_results=df_linear_total,
        allele_maps=allele_maps,
        snp=snp,
    )

    df_linear_column_renamed = df_linear_renamed.rename(
        columns={
            "coef": "Coefficient",
            "std err": "STD ERR",
            "[0.025": "0.025 CI",
            "0.975]": "0.975 CI",
        }
    )

    return df_linear_column_renamed


def compute_total_effect(
    df: pd.DataFrame,
) -> pd.DataFrame:
    intercept = df.loc["Intercept", "coef"]

    for index in df.index.drop("Intercept"):
        df.loc[index, "coef"] += intercept
        df.loc[index, "[0.025"] += intercept
        df.loc[index, "0.975]"] += intercept

    return df


def _rename_linear_regression_index(
    df_results: pd.DataFrame, allele_maps: dict[str, dict[str, str]], snp: str
) -> pd.DataFrame:
    snp_allele_map = allele_maps[snp]
    cur_mapping = {}
    for index in df_results.index:
        if index != "Intercept":
            assert snp in index, f"{snp} not in {index} of {df_results}"
        if "Intercept" in index:
            key = "REF"
        elif "[T.1]" in index:
            key = "HET"
        elif "[T.2]" in index:
            key = "ALT"
        else:
            raise ValueError()

        cur_allele = snp_allele_map[key]
        cur_mapping[index] = f"{snp} {cur_allele}"

    df_results = df_results.rename(index=cur_mapping)

    return df_results



import heapq
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Sequence

import networkx as nx
import numpy as np
import pandas as pd
from _operator import itemgetter
from aislib.misc_utils import ensure_path_exists
from eir.setup.input_setup_modules.setup_omics import read_bim
from matplotlib import pyplot as plt


def generate_interaction_snp_graph_figure(
    df: pd.DataFrame,
    bim_file_path: Path,
    plot_output_root: Path,
    df_target: pd.DataFrame,
    top_n_snps: Optional[int] = None,
    trait: Optional[str] = None,
) -> None:
    train_interaction_info = _extract_cluster_info_from_interaction_df(
        df_interactions=df, bim_file_path=bim_file_path
    )
    df_interaction_coefficients = _build_interaction_coefficient_df(
        df=df, snps=train_interaction_info.snps
    )

    df_interaction_coefficients = filter_df_interactions_for_top_n_snps(
        df_interactions=df_interaction_coefficients, top_n=top_n_snps
    )

    summed_interactions = _get_snp_interactions_from_coefficients(
        df_cluster=df_interaction_coefficients, top_n=top_n_snps
    )

    graph = build_graph_from_summed_interactions(
        summed_interactions=summed_interactions
    )

    snp_chr_map = _get_snp_chr_map(bim_file_path=bim_file_path)

    cur_fig = _get_interaction_graph_figure(
        graph=graph,
        trait=trait,
        node_color_map=snp_chr_map,
        manual_order=None,
        df_target=df_target,
    )

    cur_output_path = plot_output_root / "snp_interactions.pdf"
    ensure_path_exists(path=cur_output_path)

    cur_fig.savefig(cur_output_path, bbox_inches="tight")


def _build_interaction_coefficient_df(
    df: pd.DataFrame, snps: list[str]
) -> pd.DataFrame:
    array = np.zeros((len(snps), len(snps)))

    for row_idx, snp_1 in enumerate(snps):
        for column_idx, snp_2 in enumerate(snps):
            if snp_1 == snp_2:
                continue

            mask = (df["KEY"].str.contains(snp_1)) & (df["KEY"].str.contains(snp_2))
            df_slice = df[mask]

            if len(df_slice) == 0:
                continue

            cur_coefficient = df_slice.iloc[-1]["Coefficient"]

            array[row_idx, column_idx] = cur_coefficient

    df = pd.DataFrame(array, columns=snps, index=snps)

    return df


def _get_snp_interactions_from_coefficients(
    df_cluster: pd.DataFrame, top_n: Optional[int] = None
) -> dict:
    interactions = {}

    for snp_1 in df_cluster.index:
        for snp_2 in df_cluster.columns:
            cur_interaction = df_cluster.loc[snp_1, snp_2]

            if cur_interaction == 0.0:
                continue

            key = "---".join(sorted([snp_1, snp_2]))

            if key not in interactions:
                interactions[key] = []

            interactions[key].append(cur_interaction)

    summed_interactions = {
        key: sum(value) / len(value) for key, value in interactions.items()
    }

    if top_n and top_n < len(summed_interactions):
        summed_interactions = dict(
            heapq.nlargest(
                n=top_n,
                iterable=summed_interactions.items(),
                key=itemgetter(1),
            )
        )

    return summed_interactions


def build_graph_from_summed_interactions(
    summed_interactions: dict[str, float]
) -> nx.Graph:
    graph = nx.Graph(name="Interaction Effects")

    for key, value in summed_interactions.items():
        node_1_label, node_2_label = key.split("---")
        direction = "negative" if value < 0 else "positive"
        graph.add_weighted_edges_from(
            ebunch_to_add=[
                (
                    node_1_label,
                    node_2_label,
                    abs(value),
                )
            ],
            direction=direction,
        )

    return graph


@dataclass
class SNPInteractionClusterInfo:
    snps: list[str]
    main_effect_total: float
    interaction_effect_total: float
    percentage_interaction_effects_of_total: float


def _extract_cluster_info_from_interaction_df(
    df_interactions: pd.DataFrame, bim_file_path: Path
) -> SNPInteractionClusterInfo:
    """
    We add and trip the chromosomes when doing the order to maintain chr-pos order.
    Then we map these to rsIDs.
    """
    snps = set()

    snp_chr_map = _get_snp_chr_map(bim_file_path=bim_file_path)
    snp_pos_map = _get_snp_pos_map(bim_file_path=bim_file_path)

    main_effect_total = 0.0
    interaction_effect_total = 0.0

    for key, df_slice in df_interactions.groupby("KEY"):
        df_slice = df_slice.reset_index()
        snp_1 = df_slice.iloc[1]["allele"].split(" ")[0]
        snp_2 = df_slice.iloc[-2]["allele"].split(" ")[0]

        snp_1_chr = snp_chr_map[snp_1]
        snp_2_chr = snp_chr_map[snp_2]

        snp_1_pos = snp_pos_map[snp_1]
        snp_2_pos = snp_pos_map[snp_2]

        snp_1_w_chr_pos = f"chr{snp_1_chr}:{snp_1_pos}---{snp_1}"
        snp_2_w_chr_pos = f"chr{snp_2_chr}:{snp_2_pos}---{snp_2}"

        snps.update({snp_1_w_chr_pos, snp_2_w_chr_pos})

        main_effects = abs(df_slice.iloc[:-1]["Coefficient"]).sum()
        interaction_effect = abs(df_slice.iloc[:-1]["Coefficient"]).sum()

        main_effect_total += main_effects
        interaction_effect_total += interaction_effect

    snps = _natural_sort(tuple(snps))
    snps = tuple(i.split("---")[1] for i in snps)

    total_effects = main_effect_total + interaction_effect_total
    percentage_interaction = interaction_effect_total / total_effects

    cluster_info = SNPInteractionClusterInfo(
        snps=list(snps),
        main_effect_total=main_effect_total,
        interaction_effect_total=interaction_effect_total,
        percentage_interaction_effects_of_total=percentage_interaction,
    )

    return cluster_info


def _get_snp_pos_map(bim_file_path: Path) -> dict[str, str]:
    df = read_bim(bim_file_path=str(bim_file_path))
    chr_map = dict(zip(df["VAR_ID"], df["BP_COORD"]))
    return chr_map


def _get_snp_chr_map(bim_file_path: Path) -> dict[str, int]:
    df = read_bim(bim_file_path=str(bim_file_path))
    chr_map = dict(zip(df["VAR_ID"], df["CHR_CODE"]))

    return chr_map


def _natural_sort(sequence: Sequence) -> Sequence:
    def _convert(text: str):
        return int(text) if text.isdigit() else text.lower()

    def _alphanumeric_key(key: str) -> list:
        return [_convert(c) for c in re.split("([0-9]+)", key)]

    return sorted(sequence, key=_alphanumeric_key)


def _get_interaction_graph_figure(
    graph: nx.Graph,
    node_color_map: dict[str, int],
    trait: str,
    df_target: pd.DataFrame,
    manual_order: Sequence[str] = None,
) -> plt.Figure:
    pos = nx.spring_layout(G=graph, k=0.15, iterations=20)

    if manual_order:
        new_pos = {}
        for idx, (prev_snp, position) in enumerate(pos.items()):
            cur_snp = manual_order[idx]
            new_pos[cur_snp] = position

        pos = new_pos

    fig, ax = plt.subplots()

    cmap = plt.get_cmap("tab20")
    node_color = [cmap(node_color_map[i]) for i in graph.nodes()]

    edge_colors = []
    for u, v in graph.edges():
        cur_edge = graph[u][v]
        cur_direction = cur_edge["direction"]

        if cur_direction == "positive":
            edge_colors.append("#f54248")
        elif cur_direction == "negative":
            edge_colors.append("#42b9f5")

    nx.draw_networkx_nodes(
        G=graph,
        pos=pos,
        node_color=node_color,
        node_size=300,
    )

    nx.draw_networkx_labels(
        G=graph,
        pos=pos,
        font_size=8,
    )

    weights = scale_weights(graph=graph, df_target=df_target, trait=trait)
    nx.draw_networkx_edges(
        G=graph,
        pos=pos,
        width=weights,
        edge_color=edge_colors,
        alpha=0.8,
    )

    cur_colors = sorted(tuple(set([node_color_map[i] for i in graph.nodes])))
    for v in cur_colors:
        plt.scatter(x=[], y=[], c=[cmap(v)], label=f"{v}")

    ax.legend(
        loc="upper center",
        ncol=4,
        bbox_to_anchor=(0.5, -0.1),
        borderaxespad=0.0,
        title="Chromosome",
        frameon=False,
    )

    ax.set_title(trait)

    plt.axis("off")

    return fig


def scale_weights(
    graph: nx.Graph,
    df_target: pd.DataFrame,
    trait: str,
    scaling_factor: float = 100.0,
    max_threshold: float = 18.0,
) -> list[float]:
    target_min = df_target[trait].min()
    target_range = df_target[trait].max() - target_min

    weights = [
        (abs(graph[u][v]["weight"]) - target_min) / target_range
        for u, v in graph.edges()
    ]

    weights = [i * scaling_factor for i in weights]
    weights = [i if i < max_threshold else max_threshold for i in weights]

    return weights


def filter_df_interactions_for_top_n_snps(
    df_interactions: pd.DataFrame, top_n: int
) -> pd.DataFrame:
    df_copy = df_interactions.copy()

    top_snps = abs(df_copy).sum().sort_values().tail(top_n).index
    df_copy = df_copy.loc[df_copy.index.isin(top_snps), top_snps]

    return df_copy



from pathlib import Path
from typing import Dict, Iterable

import numpy as np
import pandas as pd
from aislib.misc_utils import ensure_path_exists
from matplotlib import pyplot as plt


def get_snp_allele_maps(
    df_acts: pd.DataFrame, rs_ids: Iterable[str]
) -> Dict[str, Dict[str, str]]:
    snp_allele_maps = {}
    for rs_id in rs_ids:
        snp_allele_maps[rs_id] = get_snp_allele_nucleotide_map(
            df_acts=df_acts, rs_id=rs_id
        )

    return snp_allele_maps


def get_snp_allele_nucleotide_map(df_acts: pd.DataFrame, rs_id: str) -> Dict[str, str]:
    nucleotide_map = {}

    ref = df_acts[df_acts["VAR_ID"] == rs_id]["REF"].item()
    alt = df_acts[df_acts["VAR_ID"] == rs_id]["ALT"].item()

    nucleotide_map["REF"] = ref * 2
    nucleotide_map["HET"] = f"{ref}{alt}"
    nucleotide_map["ALT"] = alt * 2

    return nucleotide_map


def plot_top_snps(
    df: pd.DataFrame, p_value_threshold: float, top_n: int, output_dir: Path
) -> None:
    ensure_path_exists(path=output_dir, is_folder=True)

    index_df = df.index.to_frame(index=False)
    index_df[["SNP", "Genotype"]] = index_df["allele"].str.split(" ", expand=True)
    df.reset_index(drop=True, inplace=True)
    df[["SNP", "Genotype"]] = index_df[["SNP", "Genotype"]]
    df.set_index(keys=["SNP", "Genotype"], inplace=True)

    snp_diff = df.groupby(level=0)["Coefficient"].apply(lambda x: np.abs(x - x.iloc[0]))

    snp_strength = snp_diff.groupby(level=0).sum()

    top_snps = snp_strength.nlargest(top_n).index

    for snp in top_snps:
        df_snp = df.loc[snp]

        if np.any(df_snp["p_value"] > p_value_threshold):
            continue

        if len(df_snp) < 3:
            continue

        plt.figure()
        plt.errorbar(
            x=df_snp.index.get_level_values("Genotype"),
            y=df_snp["Coefficient"],
            yerr=[
                abs(df_snp["Coefficient"] - df_snp["0.025 CI"]),
                abs(df_snp["0.975 CI"] - df_snp["Coefficient"]),
            ],
            fmt="o",
            capsize=5,
        )

        plt.title(f"SNP: {snp}")
        plt.xlabel("Genotype")
        plt.ylabel("Effect Size")
        plt.grid(True)

        plt.savefig(output_dir / f"{snp}.pdf")



from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd
import seaborn as sns
from aislib.misc_utils import ensure_path_exists

from eir_auto_gp.post_analysis.common.data_preparation import (
    FitTransformers,
    ModelData,
    SplitModelData,
)
from eir_auto_gp.post_analysis.complexity_analysis.modelling import (
    get_training_eval_iterator,
    train_and_evaluate_linear,
    train_and_evaluate_xboost,
)
from eir_auto_gp.post_analysis.complexity_analysis.viz import plot_performance

if TYPE_CHECKING:
    from eir_auto_gp.post_analysis.run_post_analysis import PostAnalysisObject

sns.set_theme(style="whitegrid")
sns.set(font_scale=1.2)


def run_complexity_analysis(post_analysis_object: "PostAnalysisObject") -> None:
    df_results = train_and_evaluate_wrapper(analysis_object=post_analysis_object)
    metrics_to_plot = _get_metric_columns(
        target_type=post_analysis_object.experiment_info.target_type
    )
    metrics_to_plot_w_average = metrics_to_plot + ["average_performance"]
    plot_output_root = (
        post_analysis_object.data_paths.analysis_output_path / "complexity/plots"
    )
    for metric in metrics_to_plot_w_average:
        plot_performance(
            df=df_results,
            output_path=plot_output_root,
            metric=metric,
        )


def train_and_evaluate_wrapper(analysis_object: "PostAnalysisObject"):
    all_results = []

    any_tabular = len(analysis_object.experiment_info.all_input_columns) > 0
    for conditions in get_training_eval_iterator(any_tabular_input=any_tabular):
        mro = convert_split_data_to_model_ready_object(
            split_model_data=analysis_object.modelling_data,
            include_genotype=conditions["include_genotype"],
            include_tabular=conditions["include_tabular"],
            one_hot_encode=conditions["one_hot_encode"],
        )

        if conditions["model_type"] == "xgboost":
            train_eval_results = train_and_evaluate_xboost(
                modelling_data=mro,
                target_type=analysis_object.experiment_info.target_type,
            )
        else:
            train_eval_results = train_and_evaluate_linear(
                modelling_data=mro,
                target_type=analysis_object.experiment_info.target_type,
            )

        df_conditions = pd.DataFrame([conditions])
        df_result = pd.concat(
            objs=[df_conditions, train_eval_results.performance],
            axis=1,
        )

        result = {
            "conditions": conditions,
            "df_predictions": train_eval_results.df_predictions,
            "df_predictions_raw": train_eval_results.df_predictions_raw,
            "performance": df_result,
        }

        all_results.append(result)

    df_all_results = process_results(
        all_results=all_results,
        target_type=analysis_object.experiment_info.target_type,
        analysis_output_path=analysis_object.data_paths.analysis_output_path,
    )

    return df_all_results


def process_results(
    all_results: list, target_type: str, analysis_output_path: Path
) -> pd.DataFrame:
    df_all_results = pd.concat(
        [r["performance"] for r in all_results],
        ignore_index=True,
    )
    numerical_columns = _get_metric_columns(target_type=target_type)

    average = df_all_results[numerical_columns].mean(axis=1)
    df_all_results["average_performance"] = average

    output_root = analysis_output_path / "complexity"
    ensure_path_exists(path=output_root, is_folder=True)
    df_all_results.to_csv(output_root / "all_results.csv", index=False)

    for i, result in enumerate(all_results):
        condition_str = "_".join(f"{k}={v}" for k, v in result["conditions"].items())
        file_name = f"predictions_{i}_{condition_str}.csv"

        numerical_output_path = output_root / "predictions/numerical" / file_name
        ensure_path_exists(path=numerical_output_path, is_folder=False)
        result["df_predictions"].to_csv(numerical_output_path)

        raw_output_path = output_root / "predictions/raw" / file_name
        ensure_path_exists(path=raw_output_path, is_folder=False)
        result["df_predictions_raw"].to_csv(raw_output_path)

    return df_all_results


def _get_metric_columns(target_type: str) -> list[str]:
    if target_type == "classification":
        return ["acc", "mcc", "roc_auc", "ap"]
    elif target_type == "regression":
        return ["r2", "pcc"]
    else:
        raise ValueError()


@dataclass()
class ModelReadyObject:
    input_train: pd.DataFrame
    target_train: pd.DataFrame

    input_val: pd.DataFrame
    target_val: pd.DataFrame

    input_test: pd.DataFrame
    target_test: pd.DataFrame

    transformers: FitTransformers


def convert_split_data_to_model_ready_object(
    split_model_data: "SplitModelData",
    include_genotype: bool = True,
    include_tabular: bool = True,
    one_hot_encode: bool = False,
) -> ModelReadyObject:
    genotype_data_combined = pd.concat(
        [
            split_model_data.train.df_genotype_input,
            split_model_data.val.df_genotype_input,
            split_model_data.test.df_genotype_input,
        ]
    )

    if one_hot_encode:
        genotype_data_combined = pd.get_dummies(
            genotype_data_combined.astype(str), drop_first=False
        )

    def get_input_data(model_data: ModelData) -> pd.DataFrame:
        data_to_include = []

        if include_genotype:
            genotype_data = genotype_data_combined.loc[
                model_data.df_genotype_input.index
            ]
            data_to_include.append(genotype_data)

        if include_tabular:
            data_to_include.append(model_data.df_tabular_input)

        return pd.concat(data_to_include, axis=1)

    input_train = get_input_data(model_data=split_model_data.train)
    target_train = split_model_data.train.df_target

    input_val = get_input_data(model_data=split_model_data.val)
    target_val = split_model_data.val.df_target

    input_test = get_input_data(model_data=split_model_data.test)
    target_test = split_model_data.test.df_target

    return ModelReadyObject(
        input_train=input_train,
        target_train=target_train,
        input_val=input_val,
        target_val=target_val,
        input_test=input_test,
        target_test=target_test,
        transformers=split_model_data.transformers,
    )



from pathlib import Path

import pandas as pd
import seaborn as sns
import xgboost
from aislib.misc_utils import ensure_path_exists
from matplotlib import pyplot as plt
from sklearn.linear_model import ElasticNetCV, LogisticRegressionCV


def assign_complexity(row):
    if row["include_tabular"] and not row["include_genotype"]:
        return 1
    elif (
        row["include_genotype"]
        and not row["include_tabular"]
        and not row["one_hot_encode"]
    ):
        return 2
    elif (
        row["include_genotype"] and not row["include_tabular"] and row["one_hot_encode"]
    ):
        return 3
    elif (
        row["include_genotype"] and row["include_tabular"] and not row["one_hot_encode"]
    ):
        return 4
    elif row["include_genotype"] and row["include_tabular"] and row["one_hot_encode"]:
        return 5


def plot_performance(df: pd.DataFrame, output_path: Path, metric: str) -> None:
    df["complexity_level"] = df.apply(assign_complexity, axis=1)

    plt.figure(figsize=(12, 8))

    sns.barplot(
        x="complexity_level",
        y=metric,
        hue="model_type",
        hue_order=["linear", "xgboost"],
        data=df,
    )
    plt.title(metric.upper())

    if not df["include_tabular"].any():
        complexity_descriptions = {
            1: "Genotype (G)",
            2: "Genotype (G), OH",
        }
        ticks_range = range(1, 3)
    else:
        complexity_descriptions = {
            1: "Tabular (T)",
            2: "Genotype (G)",
            3: "Genotype (G), OH",
            4: "G + T",
            5: "G + T, OH",
        }
        ticks_range = range(1, 6)

    plt.xticks(
        ticks=[i - 1.0 for i in ticks_range],
        labels=[complexity_descriptions[i] for i in ticks_range],
    )

    plt.xlabel("")
    plt.title("")

    best_performance = df[metric].max()
    plt.axhline(
        best_performance,
        color="r",
        linestyle="--",
        linewidth=1,
    )

    plt.legend(title="Model Type")

    plt.tight_layout()
    ensure_path_exists(path=output_path, is_folder=True)
    plt.savefig(output_path / f"{metric}_performance.pdf")


def plot_xgboost_feature_importance(
    model: xgboost.Booster, output_folder: Path
) -> None:
    xgboost.plot_importance(model)
    plt.title("Feature Importance")
    plt.tight_layout()
    plt.savefig(output_folder / "xgboost_feature_importance.pdf")


def plot_linear_coefficients(
    model: LogisticRegressionCV | ElasticNetCV,
    feature_names: list,
    output_folder: Path,
) -> None:
    coefficients = pd.Series(model.coef_, index=feature_names)
    sns.barplot(x=coefficients.values, y=coefficients.index)
    plt.xlabel("Coefficients")
    plt.ylabel("Features")
    plt.title("Feature Importance")
    plt.tight_layout()
    plt.savefig(output_folder / "linear_coefficients.pdf")



import json
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd
from aislib.misc_utils import get_logger
from eir.data_load.data_source_modules.deeplake_ops import load_deeplake_dataset
from eir.setup.input_setup_modules import setup_omics as eir_setup_omics
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, StandardScaler

from eir_auto_gp.modelling.dl_feature_selection import (
    get_dl_gwas_top_n_snp_list_df,
    get_dl_top_n_snp_list_df,
)

logger = get_logger(name=__name__)


@dataclass()
class DataPaths:
    experiment_config: Path
    train_data_path: Path
    test_data_path: Path
    train_labels_path: Path
    test_labels_path: Path
    snp_bim_path: Path
    dl_attribution_path: Path
    gwas_attribution_path: Optional[Path]
    analysis_output_path: Path


def build_data_paths(run_dir: Path) -> DataPaths:
    gwas_folder = run_dir / "feature_selection/gwas_output/"
    if gwas_folder.exists():
        gwas_attributions = next(
            i
            for i in gwas_folder.iterdir()
            if "glm.logistic" in i.name or "glm.linear" in i.name
        )
        assert gwas_attributions.exists()
    else:
        gwas_attributions = None

    config_path = run_dir / "config.json"
    config_dict = json.load(config_path.open("r"))

    data_dir = config_dict["data_output_folder"]
    if data_dir:
        data_dir = Path(data_dir)
    else:
        data_dir = run_dir / "data"

    paths = DataPaths(
        experiment_config=run_dir / "config.json",
        train_data_path=data_dir / "genotype/final/train",
        test_data_path=data_dir / "genotype/final/test",
        train_labels_path=data_dir / "tabular/final/labels_train.csv",
        test_labels_path=data_dir / "tabular/final/labels_test.csv",
        snp_bim_path=data_dir / "genotype/processed/parsed_files/data_final.bim",
        dl_attribution_path=run_dir
        / "feature_selection/dl_importance/dl_attributions.csv",
        gwas_attribution_path=gwas_attributions,
        analysis_output_path=run_dir / "analysis/post_analysis/",
    )

    return paths


@dataclass()
class ExperimentInfo:
    input_cat_columns: list[str]
    input_con_columns: list[str]
    all_input_columns: list[str]
    output_cat_columns: list[str]
    output_con_columns: list[str]
    target_type: str


@dataclass()
class ModelData:
    df_genotype_input: pd.DataFrame
    df_tabular_input: pd.DataFrame
    df_target: pd.DataFrame


def get_subset_indices_and_names(
    bim_file_path: Path,
    dl_attributions_path: Optional[Path],
    gwas_attributions_path: Optional[Path],
    top_n_snps: int,
) -> tuple[np.ndarray, list[str]]:
    df_bim = eir_setup_omics.read_bim(bim_file_path=str(bim_file_path))
    validate_file_paths(dl_path=dl_attributions_path, gwas_path=gwas_attributions_path)

    df_dl_attributions: Optional[pd.DataFrame] = None
    df_gwas_attributions: Optional[pd.DataFrame] = None

    if dl_attributions_path is not None and dl_attributions_path.exists():
        logger.info("DL attributions %s exist.", dl_attributions_path)
        df_dl_attributions = pd.read_csv(filepath_or_buffer=dl_attributions_path)

    if gwas_attributions_path is not None and gwas_attributions_path.exists():
        logger.info("GWAS attributions %s exist.", gwas_attributions_path)
        df_gwas_attributions = _read_gwas_attributions(
            gwas_attributions_path=gwas_attributions_path
        )

    match (df_dl_attributions, df_gwas_attributions):
        case (pd.DataFrame(), None):
            df_top_snps = get_dl_top_n_snp_list_df(
                df_attributions=df_dl_attributions,
                df_bim=df_bim,
                top_n_snps=top_n_snps,
            )
        case (None, pd.DataFrame()):
            df_top_snps = _get_gwas_top_n_snp_list_df(
                df_gwas=df_gwas_attributions,
                top_n_snps=top_n_snps,
            )
        case (pd.DataFrame(), pd.DataFrame()):
            df_dl_gwas = df_dl_attributions.join(other=df_gwas_attributions)
            df_top_snps = get_dl_gwas_top_n_snp_list_df(
                df_dl_gwas=df_dl_gwas,
                top_n_snps=top_n_snps,
            )
        case _:
            raise ValueError("Both DL and GWAS attributions are None.")

    top_snps_list = df_top_snps["SNP"].tolist()
    _check_bim(df_bim=df_bim, top_snps_list=top_snps_list)

    subset_indices = eir_setup_omics._setup_snp_subset_indices(
        df_bim=df_bim,
        snps_to_subset=top_snps_list,
    )

    return subset_indices, top_snps_list


def validate_file_paths(dl_path: Optional[Path], gwas_path: Optional[Path]) -> None:
    if (dl_path is None or not dl_path.exists()) and (
        gwas_path is None or not gwas_path.exists()
    ):
        raise ValueError(
            f"Neither {dl_path} nor {gwas_path} exist. "
            "Post-analysis requires at least one of these files to exist."
        )


def _get_gwas_top_n_snp_list_df(df_gwas: pd.DataFrame, top_n_snps: int) -> pd.DataFrame:
    df = df_gwas.sort_values(by="GWAS P-VALUE", ascending=True)
    df_top_n = df.iloc[:top_n_snps, :]
    df_top_n.index.name = "SNP"
    df_top_n["SNP"] = df_top_n.index
    return df_top_n


def _read_gwas_attributions(gwas_attributions_path: Path) -> pd.DataFrame:
    df_gwas = pd.read_csv(filepath_or_buffer=gwas_attributions_path, sep="\t")
    df_gwas = df_gwas.rename(columns={"ID": "VAR_ID"})
    df_gwas = df_gwas.set_index("VAR_ID")
    df_gwas = df_gwas.rename(columns={"P": "GWAS P-VALUE"})
    df_gwas = df_gwas[["GWAS P-VALUE"]]

    return df_gwas


def _check_bim(df_bim: pd.DataFrame, top_snps_list: list[str]) -> None:
    assert df_bim["VAR_ID"].nunique() == df_bim.shape[0]
    assert pd.Series(top_snps_list).isin(df_bim["VAR_ID"]).all()


def set_up_model_data(
    genotype_input_path: Path,
    labels_input_path: Path,
    experiment_info: ExperimentInfo,
    genotype_indices_to_load: np.ndarray,
    top_snps_list: list[str],
) -> ModelData:
    df_genotype_input = load_deeplake_samples_into_df(
        genotype_input_path=genotype_input_path,
        genotype_indices_to_load=genotype_indices_to_load,
        top_snps_list=top_snps_list,
    )

    df_tabular_input, df_target = load_tabular_data_into_df(
        labels_input_path=labels_input_path,
        experiment_info=experiment_info,
    )

    df_genotype_input.sort_index(inplace=True)
    df_tabular_input.sort_index(inplace=True)
    df_target.sort_index(inplace=True)

    df_target, df_genotype_input, df_tabular_input = _handle_missing_target_values(
        df_target=df_target,
        df_genotype_input=df_genotype_input,
        df_tabular_input=df_tabular_input,
    )

    assert df_genotype_input.shape[0] == df_target.shape[0] == df_tabular_input.shape[0]
    assert df_genotype_input.index.equals(df_target.index)
    assert df_genotype_input.index.equals(df_tabular_input.index)

    model_data = ModelData(
        df_genotype_input=df_genotype_input,
        df_tabular_input=df_tabular_input,
        df_target=df_target,
    )

    return model_data


def _handle_missing_target_values(
    df_target: pd.DataFrame,
    df_genotype_input: pd.DataFrame,
    df_tabular_input: pd.DataFrame,
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    logger.info("Handling missing target values, target shape: %s", df_target.shape)

    df_target = df_target.dropna()
    df_genotype_input = df_genotype_input.loc[df_target.index]
    df_tabular_input = df_tabular_input.loc[df_target.index]

    assert df_genotype_input.shape[0] == df_target.shape[0] == df_tabular_input.shape[0]

    logger.info("After dropping missing values, target shape: %s", df_target.shape)

    return df_target, df_genotype_input, df_tabular_input


def process_split_model_data(
    split_model_data: "SplitModelData", task_type: str
) -> "SplitModelData":
    transformers = _set_up_transformers(
        df_train_tab_input=split_model_data.train.df_tabular_input,
        df_train_tab_target=split_model_data.train.df_target,
        task_type=task_type,
    )

    def process_model_data(
        model_data: ModelData,
    ) -> ModelData:
        df_tabular_input_encoded = model_data.df_tabular_input
        if len(df_tabular_input_encoded.columns) > 0:
            df_tabular_input_encoded = pd.get_dummies(
                data=model_data.df_tabular_input,
                drop_first=False,
            )

            input_scaler = transformers.input_scaler
            if input_scaler is not None:
                numerical_columns = transformers.numerical_columns
                transformed_numerical = input_scaler.transform(
                    df_tabular_input_encoded[numerical_columns]
                )
                df_tabular_input_encoded[numerical_columns] = transformed_numerical

        target_data = model_data.df_target.values
        if task_type == "classification":
            label_encoder = transformers.target_label_encoder
            target_data_1d = target_data.ravel()
            target_data = label_encoder.transform(target_data_1d)

        df_target_standardized = pd.DataFrame(
            data=target_data,
            columns=model_data.df_target.columns,
            index=model_data.df_target.index,
        )

        return ModelData(
            df_genotype_input=model_data.df_genotype_input,
            df_tabular_input=df_tabular_input_encoded,
            df_target=df_target_standardized,
        )

    return SplitModelData(
        train=process_model_data(model_data=split_model_data.train),
        val=process_model_data(model_data=split_model_data.val),
        test=process_model_data(model_data=split_model_data.test),
        transformers=transformers,
    )


@dataclass()
class FitTransformers:
    input_scaler: Optional[StandardScaler]
    target_label_encoder: Optional[LabelEncoder]
    numerical_columns: list[str]


def _set_up_transformers(
    df_train_tab_input: pd.DataFrame, df_train_tab_target: pd.DataFrame, task_type: str
) -> FitTransformers:
    numerical_columns = df_train_tab_input.select_dtypes(include=[np.number]).columns

    if len(numerical_columns) > 0:
        input_scaler = StandardScaler()
        input_scaler.fit(X=df_train_tab_input[numerical_columns])
    else:
        input_scaler = None

    if task_type == "classification":
        target_transformer = LabelEncoder()
        target_transformer.fit(df_train_tab_target.values.ravel())
    elif task_type == "regression":
        target_transformer = None
    else:
        raise ValueError()

    return FitTransformers(
        input_scaler=input_scaler,
        target_label_encoder=target_transformer,
        numerical_columns=numerical_columns,
    )


def load_tabular_data_into_df(
    labels_input_path: Path,
    experiment_info: ExperimentInfo,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    all_cols = list(
        ["ID"]
        + experiment_info.input_cat_columns
        + experiment_info.input_con_columns
        + experiment_info.output_cat_columns
        + experiment_info.output_con_columns,
    )

    df_labels = pd.read_csv(
        filepath_or_buffer=labels_input_path, usecols=all_cols, index_col="ID"
    )
    df_labels.index = df_labels.index.astype(str)

    df_input = df_labels[
        experiment_info.input_cat_columns + experiment_info.input_con_columns
    ]
    df_target = df_labels[
        experiment_info.output_cat_columns + experiment_info.output_con_columns
    ]
    assert df_target.shape[1] == 1

    df_input.columns = ["COVAR_" + col for col in df_input.columns]

    return df_input, df_target


def load_deeplake_samples_into_df(
    genotype_input_path: Path,
    genotype_indices_to_load: np.ndarray,
    top_snps_list: list[str],
) -> pd.DataFrame:
    deeplake_ds = load_deeplake_dataset(data_source=str(genotype_input_path))
    ids = []
    genotype_arrays = []

    for deeplake_sample in deeplake_ds:
        sample_id = deeplake_sample["ID"].text()
        sample_genotype = deeplake_sample["genotype"].numpy()
        sample_genotype_subset = sample_genotype[:, genotype_indices_to_load]

        array_maxed = sample_genotype_subset.argmax(0)
        array_maxed[array_maxed == 3] = -1

        ids.append(sample_id)
        genotype_arrays.append(array_maxed)

    genotype_array = np.stack(genotype_arrays, axis=0)
    df_genotype = pd.DataFrame(data=genotype_array, columns=top_snps_list, index=ids)
    df_genotype.index.name = "ID"
    df_genotype.index = df_genotype.index.astype(str)

    return df_genotype


@dataclass()
class SplitModelData:
    train: ModelData
    val: ModelData
    test: ModelData

    transformers: Optional[FitTransformers] = None


def set_up_split_model_data(
    data_paths: DataPaths,
    experiment_info: ExperimentInfo,
    top_snps: int,
) -> SplitModelData:
    subset_indices, top_snps_list = get_subset_indices_and_names(
        bim_file_path=data_paths.snp_bim_path,
        dl_attributions_path=data_paths.dl_attribution_path,
        gwas_attributions_path=data_paths.gwas_attribution_path,
        top_n_snps=top_snps,
    )

    train_and_valid_data = set_up_model_data(
        genotype_input_path=data_paths.train_data_path,
        labels_input_path=data_paths.train_labels_path,
        experiment_info=experiment_info,
        genotype_indices_to_load=subset_indices,
        top_snps_list=top_snps_list,
    )

    train_data, valid_data = split_model_data_object(
        model_data=train_and_valid_data,
        test_size=0.1,
    )

    test_data = set_up_model_data(
        genotype_input_path=data_paths.test_data_path,
        labels_input_path=data_paths.test_labels_path,
        experiment_info=experiment_info,
        genotype_indices_to_load=subset_indices,
        top_snps_list=top_snps_list,
    )

    split_model_data_raw = SplitModelData(
        train=train_data,
        val=valid_data,
        test=test_data,
        transformers=None,
    )
    validate_split_model_data(split_model_data=split_model_data_raw)

    split_model_data_processed = process_split_model_data(
        split_model_data=split_model_data_raw,
        task_type=experiment_info.target_type,
    )
    validate_split_model_data(split_model_data=split_model_data_processed)

    return split_model_data_processed


def validate_split_model_data(split_model_data: "SplitModelData") -> None:
    train_ids = split_model_data.train.df_genotype_input.index
    val_ids = split_model_data.val.df_genotype_input.index
    test_ids = split_model_data.test.df_genotype_input.index

    assert len(set(train_ids) & set(val_ids)) == 0
    assert len(set(train_ids) & set(test_ids)) == 0
    assert len(set(val_ids) & set(test_ids)) == 0

    for split in [split_model_data.train, split_model_data.val, split_model_data.test]:
        assert split.df_genotype_input.index.equals(split.df_tabular_input.index)
        assert split.df_genotype_input.index.equals(split.df_target.index)


def parse_target_type(
    output_cat_columns: list[str], output_con_columns: list[str]
) -> str:
    if len(output_cat_columns) > 0:
        assert len(output_con_columns) == 0
        target_type = "classification"
    elif len(output_con_columns) > 0:
        assert len(output_cat_columns) == 0
        target_type = "regression"
    else:
        raise ValueError()

    return target_type


def split_model_data_object(
    model_data: ModelData, test_size: float = 0.2
) -> tuple[ModelData, ModelData]:
    df_input = pd.concat(
        objs=[model_data.df_genotype_input, model_data.df_tabular_input], axis=1
    )
    x_train, x_val, y_train, y_val = train_test_split(
        df_input,
        model_data.df_target,
        test_size=test_size,
        random_state=42,
    )

    train_data = ModelData(
        df_genotype_input=x_train[model_data.df_genotype_input.columns],
        df_tabular_input=x_train[model_data.df_tabular_input.columns],
        df_target=y_train,
    )

    val_data = ModelData(
        df_genotype_input=x_val[model_data.df_genotype_input.columns],
        df_tabular_input=x_val[model_data.df_tabular_input.columns],
        df_target=y_val,
    )

    return train_data, val_data


def extract_experiment_info_from_config(config_path: Path) -> ExperimentInfo:
    config_dict = json.load(config_path.open("r"))

    input_cat_columns = config_dict["input_cat_columns"]
    input_con_columns = config_dict["input_con_columns"]
    all_input_columns = input_cat_columns + input_con_columns

    output_cat_columns = config_dict["output_cat_columns"]
    output_con_columns = config_dict["output_con_columns"]

    target_type = parse_target_type(
        output_cat_columns=output_cat_columns, output_con_columns=output_con_columns
    )

    experiment_info = ExperimentInfo(
        input_cat_columns=input_cat_columns,
        input_con_columns=input_con_columns,
        all_input_columns=all_input_columns,
        output_cat_columns=output_cat_columns,
        output_con_columns=output_con_columns,
        target_type=target_type,
    )

    return experiment_info



import warnings
from io import StringIO
from itertools import combinations
from pathlib import Path

import numpy as np
import pandas as pd
from aislib.misc_utils import get_logger
from eir.setup.input_setup_modules.setup_omics import read_bim
from joblib import Parallel, delayed
from statsmodels.base.model import LikelihoodModelResults
from statsmodels.tools.sm_exceptions import ConvergenceWarning

from eir_auto_gp.post_analysis.effect_analysis.genotype_effects import (
    get_snp_allele_maps,
    get_statsmodels_fit_function,
)

logger = get_logger(name=__name__)

warnings.filterwarnings(action="ignore", category=ConvergenceWarning)
warnings.filterwarnings(action="ignore", category=RuntimeWarning)


def get_interaction_effects(
    df_genotype: pd.DataFrame,
    df_target: pd.DataFrame,
    bim_file: Path,
    target_type: str,
) -> pd.DataFrame:
    target_name = df_target.columns[0]
    df_combined = pd.concat(objs=[df_target, df_genotype], axis=1)

    df_bim = read_bim(bim_file_path=str(bim_file))
    snp_ids = df_genotype.columns
    allele_maps = get_snp_allele_maps(df_bim=df_bim, snp_ids=snp_ids)

    df_results = compute_interactions(
        df=df_combined,
        target_name=target_name,
        allele_maps=allele_maps,
        target_type=target_type,
        p_threshold="auto",
    )

    return df_results


def compute_interactions(
    df: pd.DataFrame,
    target_name: str,
    allele_maps: dict[str, dict[str, str]],
    target_type: str,
    p_threshold: str | float = "auto",
) -> pd.DataFrame:
    logger.info(
        "Gathering all interaction allele effect results for '%s'.",
        target_name,
    )

    snps = [i for i in df.columns if i != target_name]

    if p_threshold == "auto":
        p_threshold = 0.05 / (len(snps) ** 2)

        logger.info(
            "p value threshold set to %f due to 'auto' option and %d SNPs.",
            p_threshold,
            len(snps),
        )

    snp_iter = combinations(iterable=snps, r=2)

    parallel_worker = Parallel(n_jobs=-1)
    all_results = parallel_worker(
        delayed(_compute_interaction_snp_effect_wrapper)(
            df=df,
            target_name=target_name,
            allele_maps=allele_maps,
            snp_1=snp_1,
            snp_2=snp_2,
            p_threshold=p_threshold,
            target_type=target_type,
        )
        for snp_1, snp_2 in snp_iter
    )
    all_results = [i for i in all_results if i is not None]

    if len(all_results) == 0:
        return pd.DataFrame()

    return pd.concat(all_results)


def _compute_interaction_snp_effect_wrapper(
    df: pd.DataFrame,
    target_name: str,
    snp_1: str,
    snp_2: str,
    p_threshold: float,
    allele_maps: dict[str, dict[str, str]],
    target_type: str,
) -> pd.DataFrame | None:
    formula = build_multiplicative_interaction_snp_formula(
        target=target_name,
        snp_1=snp_1,
        snp_2=snp_2,
    )

    df_cur = df[[target_name, snp_1, snp_2]]
    df_cur_no_na = df_cur[df_cur[snp_1] != -1]
    df_cur_no_na = df_cur_no_na[df_cur_no_na[snp_2] != -1]

    fit_func = get_statsmodels_fit_function(target_type=target_type)

    try:
        result = fit_func(formula=formula, data=df_cur_no_na).fit(disp=0)

        if hasattr(result, "mle_retvals"):
            if not result.mle_retvals["converged"]:
                raise RuntimeError("Model did not converge.")
        elif not np.all(np.isfinite(result.normalized_cov_params)):
            raise RuntimeError("Model did not converge.")

        result_summary = result.summary2().tables[1]
        if not np.all(np.isfinite(result_summary)):
            raise RuntimeError("Model resulted in non-finite values")

    except Exception as exception:
        logger.info(
            "Failed OLS on '%s' and '%s' due to exception '%s'. Skipping.",
            snp_1,
            snp_2,
            exception,
        )
        return None

    p_value = _extract_p_value_from_interaction_results(results=result)
    if p_value > p_threshold:
        return None

    try:
        df_result = build_df_from_interaction_results(
            results=result, allele_maps=allele_maps, snp_1=snp_1, snp_2=snp_2
        )
        return df_result
    except Exception as e:
        logger.info(
            "Failed parsing interaction results on '%s' and '%s' due to exception "
            "'%s'. Skipping.",
            snp_1,
            snp_2,
            e,
        )
        return None


def _extract_p_value_from_interaction_results(
    results: LikelihoodModelResults,
) -> float:
    p_values = results.pvalues
    last_row_name = p_values.index[-1]
    assert ":" in last_row_name
    p_value = p_values.iloc[-1]
    return p_value


def build_multiplicative_interaction_snp_formula(
    target: str,
    snp_1: str,
    snp_2: str,
) -> str:
    return (
        f"Q('{target}') ~ C(Q('{snp_1}')) + C(Q('{snp_2}')) + Q('{snp_1}'):Q('{snp_2}')"
    )


def build_df_from_interaction_results(
    results: LikelihoodModelResults,
    allele_maps: dict[str, dict[str, str]],
    snp_1: str,
    snp_2: str,
) -> pd.DataFrame:
    results_as_html = results.summary().tables[1].as_html()
    html_buffer = StringIO(results_as_html)
    df_linear = pd.read_html(io=html_buffer, header=0, index_col=0)[0]
    df_linear.index.name = "allele"
    df_linear.index = df_linear.index.str.replace(r"C\(Q\(", "", regex=True)
    df_linear.index = df_linear.index.str.replace(r"Q\(", "", regex=True)
    df_linear.index = df_linear.index.str.replace(r"\)", "", regex=True)
    df_linear.index = df_linear.index.str.replace(r"'", "", regex=True)

    df_linear_renamed = _rename_interaction_regression_index(
        df_results=df_linear, allele_maps=allele_maps, snp=snp_1
    )

    df_linear_renamed = _rename_interaction_regression_index(
        df_results=df_linear_renamed, allele_maps=allele_maps, snp=snp_2
    )

    df_linear_renamed["KEY"] = f"{snp_1}:{snp_2}"

    df_linear_column_renamed = df_linear_renamed.rename(
        columns={
            "coef": "Coefficient",
            "std err": "STD ERR",
            "[0.025": "0.025 CI",
            "0.975]": "0.975 CI",
        }
    )

    return df_linear_column_renamed


def _rename_interaction_regression_index(
    df_results: pd.DataFrame, allele_maps: dict[str, dict[str, str]], snp: str
):
    snp_allele_map = allele_maps[snp]
    cur_mapping = {}
    for index in df_results.index:
        if "[T.1]" in index and snp in index:
            key = "HET"
        elif "[T.2]" in index and snp in index:
            key = "ALT"
        else:
            continue

        cur_allele = snp_allele_map[key]
        cur_mapping[index] = f"{snp} {cur_allele}"

    df_results = df_results.rename(index=cur_mapping)

    return df_results



import argparse
import json
import shutil
from argparse import RawTextHelpFormatter
from copy import copy
from pathlib import Path
from typing import Any, Dict, Optional, Sequence

import luigi
import pandas as pd
from aislib.misc_utils import ensure_path_exists

from eir_auto_gp.analysis.run_analysis import RunAnalysisWrapper
from eir_auto_gp.preprocess.converge import ParseDataWrapper
from eir_auto_gp.preprocess.gwas_pre_selection import validate_geno_data_path
from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


def get_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(formatter_class=RawTextHelpFormatter)

    parser.add_argument(
        "--genotype_data_path",
        type=str,
        required=True,
        help="Root path to raw genotype data to be processed\n"
        "(e.g., containing my_data.bed, my_data.fam, my_data.bim).\n"
        "For this example, this parameter should be\n"
        "'/path/to/raw/genotype/data/'.\n"
        "Note that the file names are not included in this path,\n"
        "only the root folder. The file names are inferred, and\n"
        "*only one* set of files is expected.",
    )

    parser.add_argument(
        "--genotype_processing_chunk_size",
        type=int,
        default=1000,
        help="Chunk size for processing genotype data. Inreasing"
        "this value will increase the memory usage, but will"
        "likely speed up the processing.",
    )

    parser.add_argument(
        "--label_file_path",
        type=str,
        required=True,
        help="File path to label file with tabular inputs and labels to predict.",
    )

    parser.add_argument(
        "--only_data",
        action="store_true",
        required=False,
        help="If this flag is set, only the data processing step will be run.",
    )

    parser.add_argument(
        "--global_output_folder",
        type=str,
        required=False,
        help="Common root folder to save data, feature selection and modelling results"
        " in.",
    )

    parser.add_argument(
        "--data_output_folder",
        type=str,
        required=False,
        help="Folder to save the processed data in and also to read the data from"
        "if it already exists.",
    )

    parser.add_argument(
        "--feature_selection_output_folder",
        type=str,
        required=False,
        help="Folder to save feature selection results in.",
    )

    parser.add_argument(
        "--modelling_output_folder",
        type=str,
        required=False,
        help="Folder to save modelling results in.",
    )

    parser.add_argument(
        "--analysis_output_folder",
        type=str,
        required=False,
        help="Folder to save analysis results in.",
    )

    parser.add_argument(
        "--output_name",
        type=str,
        default="genotype",
        help="Name used for dataset.",
    )

    parser.add_argument(
        "--pre_split_folder",
        type=str,
        required=False,
        help="If there is a pre-split folder, this will be used to\n"
        "split the data into train/val and test sets. If not,\n"
        "the data will be split randomly.\n"
        "The folder should contain the following files:\n"
        "  - train_ids.txt: List of sample IDs to use for training.\n"
        "  - test_ids.txt: List of sample IDs to use for testing.\n"
        "  - (Optional): valid_ids.txt: List of sample IDs to use for validation.\n"
        "If this option is not specified, the data will be split randomly\n"
        "into 90/10 (train+val)/test sets.",
    )

    parser.add_argument(
        "--freeze_validation_set",
        help="If this flag is set, the validation set will be frozen\n"
        "and not changed between DL training folds.\n"
        "This only has an effect if the validation set is not specified\n"
        "in as a valid_ids.txt in file the pre_split_folder.\n"
        "If this flag is not set, the validation set will be randomly\n"
        "selected from the training set each time in each DL training run fold.\n"
        "This also has an effect when GWAS is used in feature selection.\n"
        "If the validation set is not specified manually or this flag is set,\n"
        "the GWAS will be performed on the training *and* validation set.\n"
        "This might potentially inflate the results on the validation set,\n"
        "particularly if the dataset is small. To turn off this behavior,\n"
        "you can use the --no-freeze_validation_set flag.",
        default=True,
        action="store_true",
    )

    parser.add_argument(
        "--no-freeze_validation_set",
        dest="freeze_validation_set",
        action="store_false",
    )

    parser.add_argument(
        "--feature_selection",
        default="gwas->dl",
        choices=["dl", "gwas", "gwas->dl", "dl+gwas", None],
        required=False,
        help="What kind of feature selection strategy to use for SNP selection:\n"
        "  - If None, no feature selection is performed.\n"
        "  - If 'dl', feature selection is performed using DL feature importance,\n"
        "    and the top SNPs are selected iteratively using Bayesian optimization.\n"
        "  - If 'gwas', feature selection is performed using GWAS p-values,\n"
        "    as specified by the --gwas_p_value_threshold parameter.\n"
        "  - If 'gwas->dl', feature selection is first performed using GWAS p-values,\n"
        "    and then the top SNPs are selected iteratively using the DL "
        "importance method,\n"
        "    but only on the SNPs under the GWAS threshold.",
    )

    parser.add_argument(
        "--n_dl_feature_selection_setup_folds",
        type=int,
        default=3,
        required=False,
        help="How many folds to run DL attribution calculation on genotype data\n"
        "before using results from attributions for feature selection.\n"
        "Applicable only if using 'dl' or 'gwas->dl' feature_selection options.",
    )

    parser.add_argument(
        "--gwas_p_value_threshold",
        type=float,
        required=False,
        default=1e-04,
        help="GWAS p-value threshold for filtering if using 'gwas' or 'gwas->dl'\n"
        "feature_selection options.",
    )

    parser.add_argument(
        "--folds",
        type=str,
        default="0-5",
        help="Training runs / folds to run, can be a single fold (e.g. 0),\n"
        "a range of folds (e.g. 0-5), or a comma-separated list of \n"
        "folds (e.g. 0,1,2,3,4,5).",
    )

    parser.add_argument(
        "--input_cat_columns",
        nargs="*",
        type=str,
        default=[],
        help="List of categorical columns to use as input.",
    )

    parser.add_argument(
        "--input_con_columns",
        nargs="*",
        type=str,
        default=[],
        help="List of continuous columns to use as input.",
    )

    parser.add_argument(
        "--output_cat_columns",
        nargs="*",
        type=str,
        default=[],
        help="List of categorical columns to use as output.",
    )

    parser.add_argument(
        "--output_con_columns",
        nargs="*",
        type=str,
        default=[],
        help="List of continuous columns to use as output.",
    )

    parser.add_argument(
        "--do_test",
        action="store_true",
        help="Whether to run test set prediction.",
    )

    return parser


def get_cl_args(parser: argparse.ArgumentParser) -> argparse.Namespace:
    cl_args = parser.parse_args()

    return cl_args


def validate_label_file(
    label_file_path: str,
    input_cat_columns: list[str],
    input_con_columns: list[str],
    output_cat_columns: list[str],
    output_con_columns: list[str],
) -> None:
    if not Path(label_file_path).exists():
        raise ValueError(
            f"Label file path {label_file_path} is invalid. "
            f"Expected to find {label_file_path}."
        )

    columns = pd.read_csv(label_file_path, nrows=1).columns
    if "ID" not in columns:
        raise ValueError(
            f"Label file path {label_file_path} is invalid. "
            f"Expected to find 'ID' column."
        )

    all_columns = set(
        input_cat_columns + input_con_columns + output_cat_columns + output_con_columns
    )
    missing_columns = all_columns - set(columns)
    if len(missing_columns) > 0:
        raise ValueError(
            f"Label file path {label_file_path} is invalid. "
            f"Expected to find columns {missing_columns}."
        )


def validate_targets(
    output_con_columns: list[str], output_cat_columns: list[str], only_data: bool
) -> None:
    if not only_data and len(output_con_columns) == 0 and len(output_cat_columns) == 0:
        raise ValueError(
            "At least one output column must be specified as continuous or categorical."
        )

    if len(output_con_columns) + len(output_cat_columns) > 1:
        raise ValueError(
            "Currently only one target column per run is supported. Got "
            f"{output_con_columns} continuous and {output_cat_columns} "
            "categorical target columns."
        )


def validate_plink2_exists_in_path() -> None:
    if shutil.which("plink2") is None:
        raise ValueError(
            "plink2 is not installed or not in the path. "
            "Please install plink2 and try again."
        )


def validate_pre_split_folder(pre_split_folder: Optional[str]) -> None:
    if not pre_split_folder:
        return

    ids = {}
    for file in ["train_ids.txt", "valid_ids.txt", "test_ids.txt"]:
        if not Path(pre_split_folder, file).exists():
            continue
        with open(Path(pre_split_folder, file), "r") as f:
            ids[file] = set(f.read().splitlines())

    if len(ids) == 0:
        raise ValueError(
            f"Pre-split folder {pre_split_folder} is invalid. "
            f"Expected to find at least train_ids.txt and test_ids.txt."
        )

    assert "train_ids.txt" in ids
    assert "test_ids.txt" in ids

    if "valid_ids.txt" in ids:
        assert ids["valid_ids.txt"].isdisjoint(ids["train_ids.txt"])
        assert ids["valid_ids.txt"].isdisjoint(ids["test_ids.txt"])

    assert ids["train_ids.txt"].isdisjoint(ids["test_ids.txt"])


def run(cl_args: argparse.Namespace) -> None:
    validate_geno_data_path(geno_data_path=cl_args.genotype_data_path)
    validate_label_file(
        label_file_path=cl_args.label_file_path,
        input_cat_columns=cl_args.input_cat_columns,
        input_con_columns=cl_args.input_con_columns,
        output_cat_columns=cl_args.output_cat_columns,
        output_con_columns=cl_args.output_con_columns,
    )
    validate_targets(
        output_con_columns=cl_args.output_con_columns,
        output_cat_columns=cl_args.output_cat_columns,
        only_data=cl_args.only_data,
    )
    validate_plink2_exists_in_path()
    validate_pre_split_folder(pre_split_folder=cl_args.pre_split_folder)

    cl_args = parse_output_folders(cl_args=cl_args)
    cl_args = _add_pre_split_folder_if_present(cl_args=cl_args)

    data_config = build_data_config(cl_args=cl_args)
    feature_selection_config = build_feature_selection_config(cl_args=cl_args)
    modelling_config = build_modelling_config(cl_args=cl_args)
    analysis_config = build_analysis_config(cl_args=cl_args)
    root_task = get_root_task(
        folds=cl_args.folds,
        data_config=data_config,
        feature_selection_config=feature_selection_config,
        modelling_config=modelling_config,
        analysis_config=analysis_config,
    )

    luigi.build(
        tasks=[root_task],
        workers=1,
        local_scheduler=True,
    )


def get_root_task(
    data_config: Dict,
    folds: int,
    feature_selection_config: Dict,
    modelling_config: Dict,
    analysis_config: Dict,
) -> RunAnalysisWrapper | ParseDataWrapper:
    if data_config.get("only_data"):
        return ParseDataWrapper(data_config=data_config)

    return RunAnalysisWrapper(
        folds=folds,
        data_config=data_config,
        feature_selection_config=feature_selection_config,
        modelling_config=modelling_config,
        analysis_config=analysis_config,
    )


def main():
    parser = get_argument_parser()
    cl_args = get_cl_args(parser=parser)
    store_experiment_config(cl_args=cl_args)

    run(cl_args=cl_args)


def store_experiment_config(
    cl_args: argparse.Namespace,
) -> None:
    config_dict = vars(cl_args)

    if cl_args.global_output_folder is None:
        output_folder = Path(cl_args.modelling_output_folder).parent
    else:
        output_folder = Path(cl_args.global_output_folder)

    ensure_path_exists(path=output_folder, is_folder=True)
    output_path = output_folder / "config.json"

    if output_path.exists():
        logger.warning(
            f"Output config file {output_path} already exists. Overwriting it."
        )

    with open(output_path, "w") as f:
        json.dump(config_dict, f, indent=4)


def parse_output_folders(cl_args: argparse.Namespace) -> argparse.Namespace:
    cl_args_copy = copy(cl_args)
    if cl_args_copy.global_output_folder:
        gof = cl_args_copy.global_output_folder.rstrip("/")
        cl_args_copy.data_output_folder = gof + "/data"
        cl_args_copy.feature_selection_output_folder = gof + "/feature_selection"
        cl_args_copy.modelling_output_folder = gof + "/modelling"
        cl_args_copy.analysis_output_folder = gof + "/analysis"
    else:
        if not cl_args_copy.data_output_folder:
            raise ValueError(
                "Missing data output folder. "
                "Either a global output folder or a "
                "data output folder must be provided."
            )
        if not cl_args_copy.feature_selection_output_folder:
            raise ValueError(
                "Missing feature selection output folder. "
                "Either a global output folder or a "
                "feature selection output folder must be provided."
            )
        if not cl_args_copy.modelling_output_folder:
            raise ValueError(
                "Missing modelling output folder. "
                "Either a global output folder or a "
                "modelling output folder must be provided."
            )
        if not cl_args_copy.analysis_output_folder:
            raise ValueError(
                "Missing analysis output folder. "
                "Either a global output folder or a "
                "analysis output folder must be provided."
            )

    return cl_args_copy


def _add_pre_split_folder_if_present(cl_args: argparse.Namespace) -> argparse.Namespace:
    cl_args_copy = copy(cl_args)
    genotype_path = Path(cl_args_copy.genotype_data_path)

    id_path = genotype_path / "ids"
    if id_path.exists():
        cl_args_copy.pre_split_folder = str(genotype_path / "ids")

        found_files = [
            i.name
            for i in id_path.iterdir()
            if i.is_file() and i.name.endswith(".txt") and "_plink" not in i.name
        ]

        logger.info(
            f"Found pre-split folder '{cl_args_copy.pre_split_folder}'. "
            f"in root genotype folder. "
            f"Using files: {found_files} for respective splits."
        )

    return cl_args_copy


def build_data_config(cl_args: argparse.Namespace) -> Dict[str, Any]:
    data_keys = [
        "genotype_data_path",
        "label_file_path",
        "data_output_folder",
        "output_name",
        "only_data",
        "pre_split_folder",
        "freeze_validation_set",
        "genotype_processing_chunk_size",
    ]

    base = extract_from_namespace(namespace=cl_args, keys=data_keys)
    base["output_format"] = "deeplake"

    return base


def build_feature_selection_config(cl_args: argparse.Namespace) -> Dict[str, Any]:
    feature_selection_keys = [
        "feature_selection_output_folder",
        "feature_selection",
        "n_dl_feature_selection_setup_folds",
        "gwas_p_value_threshold",
    ]

    return extract_from_namespace(namespace=cl_args, keys=feature_selection_keys)


def build_modelling_config(cl_args: argparse.Namespace) -> Dict[str, Any]:
    modelling_keys = [
        "modelling_output_folder",
        "input_cat_columns",
        "input_con_columns",
        "output_cat_columns",
        "output_con_columns",
        "do_test",
    ]

    return extract_from_namespace(namespace=cl_args, keys=modelling_keys)


def build_analysis_config(cl_args: argparse.Namespace) -> Dict[str, Any]:
    analysis_keys = [
        "analysis_output_folder",
    ]

    return extract_from_namespace(namespace=cl_args, keys=analysis_keys)


def extract_from_namespace(
    namespace: argparse.Namespace, keys: Sequence[str]
) -> Dict[str, Any]:
    return {key: getattr(namespace, key) for key in keys}


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING

from aislib.misc_utils import ensure_path_exists, get_logger

from eir_auto_gp.post_analysis.effect_analysis.genotype_effects import (
    get_allele_effects,
)
from eir_auto_gp.post_analysis.effect_analysis.interaction_effects import (
    get_interaction_effects,
)
from eir_auto_gp.post_analysis.effect_analysis.viz_genotype_effects import plot_top_snps
from eir_auto_gp.post_analysis.effect_analysis.viz_interaction_effects import (
    generate_interaction_snp_graph_figure,
)
from eir_auto_gp.post_analysis.run_complexity_analysis import (
    convert_split_data_to_model_ready_object,
)

if TYPE_CHECKING:
    from eir_auto_gp.post_analysis.run_post_analysis import PostAnalysisObject

logger = get_logger(name=__name__)


def run_effect_analysis(post_analysis_object: "PostAnalysisObject") -> None:
    mro_genotype = convert_split_data_to_model_ready_object(
        split_model_data=post_analysis_object.modelling_data,
        include_genotype=True,
        include_tabular=False,
        one_hot_encode=False,
    )

    output_root = (
        post_analysis_object.data_paths.analysis_output_path / "effect_analysis"
    )
    ensure_path_exists(path=output_root, is_folder=True)

    df_allele_effects = get_allele_effects(
        df_genotype=mro_genotype.input_train,
        df_target=mro_genotype.target_train,
        bim_file=post_analysis_object.data_paths.snp_bim_path,
        target_type=post_analysis_object.experiment_info.target_type,
    )
    df_allele_effects.to_csv(output_root / "allele_effects.csv")

    plot_top_snps(
        df=df_allele_effects,
        p_value_threshold=0.05,
        top_n=10,
        output_dir=output_root / "top_snps",
    )

    df_interaction_effects = get_interaction_effects(
        df_genotype=mro_genotype.input_train,
        df_target=mro_genotype.target_train,
        bim_file=post_analysis_object.data_paths.snp_bim_path,
        target_type=post_analysis_object.experiment_info.target_type,
    )
    df_interaction_effects.to_csv(output_root / "interaction_effects.csv")

    if len(df_interaction_effects) > 0:
        trait_name = mro_genotype.target_train.columns[0]
        generate_interaction_snp_graph_figure(
            df=df_interaction_effects,
            bim_file_path=post_analysis_object.data_paths.snp_bim_path,
            df_target=mro_genotype.target_train,
            trait=trait_name,
            plot_output_root=output_root / "interaction_graphs",
            top_n_snps=10,
        )



import math
import warnings
from pathlib import Path
from shutil import copyfile
from typing import Dict, Optional, Sequence, Tuple

warnings.filterwarnings("ignore", message=".*newer version of deeplake.*")

import deeplake
import eir.data_load.label_setup
import luigi
import numpy as np
import pandas as pd
from aislib.misc_utils import ensure_path_exists
from eir.data_load.data_source_modules.deeplake_ops import load_deeplake_dataset

from eir_auto_gp.preprocess.genotype import FinalizeGenotypeParsing
from eir_auto_gp.preprocess.tabular import ParseLabelFile
from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


@deeplake.compute
def _populate_deeplake_ds(
    id_array_tuple: Tuple[str, np.ndarray],
    deeplake_ds: deeplake.Dataset,
    output_name: str,
) -> None:
    """
    This would have been nice to have work with e.g. something like:

    with ds:
        parallel_function = _populate_deeplake_ds(
            output_name=output_name,
            argmax_replace_map=argmax_replace_map,
        )
        parallel_function.eval(data_in=id_array_generator, data_out=ds, num_workers=10)

    However, deeplake doesn't support having generators for eval yet.

    We could maybe do something like this for now, but needs to be tested:

    cur_chunk = []
    for index, (ids, arrays) in enumerate(id_array_generator):
        cur_chunk.append((ids, arrays))
        if len(cur_chunk) == parallel_chunk_size:
            with ds:
                parallel_function = _populate_deeplake_ds(
                    output_name=output_name,
                    argmax_replace_map=argmax_replace_map,
                )
                parallel_function.eval(
                    data_in=cur_chunk, data_out=ds, num_workers=10
                )

            cur_chunk = []

    if cur_chunk:
        with ds:
            parallel_function = _populate_deeplake_ds(
                output_name=output_name,
                argmax_replace_map=argmax_replace_map,
            )
            parallel_function.eval(data_in=cur_chunk, data_out=ds, num_workers=10)
    """

    id_, array = id_array_tuple

    sample = {
        "ID": id_,
        output_name: array,
    }

    deeplake_ds.append(sample)


class CommonSplitIntoTestSet(luigi.Task):
    genotype_data_path = luigi.Parameter()
    label_file_path = luigi.Parameter()
    data_output_folder = luigi.Parameter()
    output_format = luigi.Parameter()
    output_name = luigi.Parameter()
    pre_split_folder = luigi.Parameter()
    freeze_validation_set = luigi.BoolParameter()
    only_data = luigi.BoolParameter()
    genotype_processing_chunk_size = luigi.IntParameter()

    def requires(self):
        """
        Will require 3 tasks:
            - genotype processing,
            - label file processing.

        Will gather IDs from all of these, then generate a split from that.

        One option for configuration is to allow custom overwriting of QC command.
        """

        geno_task = FinalizeGenotypeParsing(
            raw_data_path=self.genotype_data_path,
            output_folder=str(self.data_output_folder) + "/genotype",
            output_name=self.output_name,
            output_format=self.output_format,
            genotype_processing_chunk_size=self.genotype_processing_chunk_size,
        )

        label_file_task = ParseLabelFile(
            output_folder=self.data_output_folder,
            label_file_path=self.label_file_path,
        )
        return {
            "genotype": geno_task,
            "label_file": label_file_task,
        }

    def run(self):
        inputs = self.input()

        genotype_path = Path(str(inputs["genotype"][1].path)) / "genotype"
        assert genotype_path.exists()
        label_file_path = Path(str(inputs["label_file"].path))

        output_root = Path(str(self.data_output_folder))

        common_ids_to_keep = _gather_all_ids(
            genotype_path=genotype_path, label_file=label_file_path
        )

        train_ids, valid_ids, test_ids = _id_setup_wrapper(
            output_root=output_root,
            pre_split_folder=self.pre_split_folder,
            common_ids_to_keep=common_ids_to_keep,
            freeze_validation_set=self.freeze_validation_set,
        )
        train_and_valid_ids = train_ids + valid_ids

        _split_csv_into_train_and_test(
            train_ids=train_and_valid_ids,
            test_ids=test_ids,
            source=label_file_path,
            destination=output_root / "tabular" / "final",
        )

        _split_deeplake_ds_into_train_and_test(
            train_ids=train_and_valid_ids,
            test_ids=test_ids,
            source=genotype_path,
            destination=output_root / "genotype" / "final",
        )

    def output(self):
        output_root = Path(str(self.data_output_folder))
        outputs = {}

        for split in ("train", "test"):
            cur_target = luigi.LocalTarget(str(output_root / "genotype/final" / split))
            outputs[f"{split}_genotype"] = cur_target

            cur_tabular_path = (
                output_root / "tabular/final" / f"labels_{split}"
            ).with_suffix(".csv")
            cur_tabular_target = luigi.LocalTarget(str(cur_tabular_path))
            outputs[f"{split}_tabular"] = cur_tabular_target

        return outputs


def _id_setup_wrapper(
    output_root: Path,
    common_ids_to_keep: Sequence[str],
    freeze_validation_set: bool,
    pre_split_folder: Optional[str] = None,
) -> tuple[list[str], list[str], list[str]]:
    valid_path = None
    valid_path_exists = False

    if pre_split_folder is not None:
        pre_split_folder = Path(pre_split_folder)
        logger.info("Using pre-split folder: %s", pre_split_folder)

        train_path = pre_split_folder / "train_ids.txt"
        if not train_path.exists():
            raise FileNotFoundError(f"Could not find train IDs file: {train_path}")

        test_path = pre_split_folder / "test_ids.txt"
        if not test_path.exists():
            raise FileNotFoundError(f"Could not find test IDs file: {test_path}")

        train_ids = (
            pd.read_csv(train_path, header=None).astype(str).squeeze("columns").tolist()
        )
        check_extra_ids(
            ids_to_check=train_ids,
            id_set_name="train",
            common_ids=common_ids_to_keep,
        )

        test_ids = (
            pd.read_csv(test_path, header=None).astype(str).squeeze("columns").tolist()
        )
        check_extra_ids(
            ids_to_check=test_ids,
            id_set_name="test",
            common_ids=common_ids_to_keep,
        )

        valid_path = pre_split_folder / "valid_ids.txt"
        valid_path_exists = valid_path.exists()

    else:
        logger.info("Generating train+valid and test IDs.")

        train_ids, test_ids = _split_ids(all_ids=common_ids_to_keep)

    valid_ids = []
    if freeze_validation_set:
        if not valid_path_exists:
            logger.info("Creating new frozen validation set.")

            batch_size = get_batch_size(samples_per_epoch=len(train_ids))
            valid_size = get_dynamic_valid_size(
                num_samples_per_epoch=len(train_ids),
                minimum=batch_size,
            )
            train_ids, valid_ids = _split_ids(
                all_ids=train_ids, valid_or_test_size=valid_size
            )
        else:
            logger.info("Using frozen validation set from %s.", valid_path)
            valid_ids = (
                pd.read_csv(valid_path, header=None)
                .astype(str)
                .squeeze("columns")
                .tolist()
            )
            check_extra_ids(
                ids_to_check=valid_ids,
                id_set_name="valid",
                common_ids=common_ids_to_keep,
            )

        _save_ids_to_text_file(
            ids=valid_ids, path=output_root / "ids" / "valid_ids.txt"
        )

    _save_ids_to_text_file(ids=train_ids, path=output_root / "ids" / "train_ids.txt")
    _save_ids_to_text_file(ids=test_ids, path=output_root / "ids" / "test_ids.txt")

    logger.info("Train and valid IDs: %d", len(train_ids))
    logger.info("Test IDs: %d", len(test_ids))

    check_missing_ids(
        train_ids=train_ids,
        valid_ids=valid_ids,
        test_ids=test_ids,
        common_ids=common_ids_to_keep,
    )

    return train_ids, valid_ids, test_ids


def check_extra_ids(
    ids_to_check: Sequence[str],
    id_set_name: str,
    common_ids: Sequence[str],
    preview_limit: int = 10,
) -> None:
    common_ids_set = set(common_ids)

    extra_ids = list(set(ids_to_check) - common_ids_set)
    if extra_ids:
        preview = extra_ids[:preview_limit]
        raise ValueError(
            f"{id_set_name} contains IDs not in common IDs from "
            f"genotype data and label file."
            f"Preview of extra IDs: {preview}."
            f"Please check that the IDs in {id_set_name} "
            f"are available in both genotype data and label file."
        )


def check_missing_ids(
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    test_ids: Sequence[str],
    common_ids: Sequence[str],
    preview_limit: int = 10,
) -> None:
    common_ids_set = set(common_ids)

    combined_ids = set(train_ids) | set(valid_ids) | set(test_ids)
    missing_ids = list(common_ids_set - combined_ids)
    if missing_ids:
        preview = missing_ids[:preview_limit]
        logger.warning(
            f"Some common IDs are missing in the final sets. "
            f"Preview of missing IDs: {preview}"
        )


def _save_ids_to_text_file(ids: Sequence[str], path: Path) -> None:
    ensure_path_exists(path=path, is_folder=False)
    with open(path, "w") as f:
        for cur_id in ids:
            f.write(f"{cur_id}\n")


def _gather_all_ids(
    genotype_path: Path, label_file: Path, filter_common: bool = True
) -> Sequence[str]:
    genotype_ids = set(
        eir.data_load.label_setup.gather_ids_from_data_source(data_source=genotype_path)
    )
    logger.info(
        "Gathered %d IDs from genotype data: ", len(genotype_ids), genotype_path
    )

    label_file_ids = set(gather_ids_from_csv_file(file_path=label_file))
    logger.info("Gathered %d IDs from label file: %s", len(label_file_ids), label_file)

    all_ids = set().union(genotype_ids, label_file_ids)
    if filter_common:
        common_ids = label_file_ids.intersection(genotype_ids)
        logger.info(
            "Keeping %d common IDs among %d total (difference: %d).",
            len(common_ids),
            len(all_ids),
            len(all_ids) - len(common_ids),
        )
        return list(common_ids)

    return list(all_ids)


def gather_ids_from_csv_file(file_path: Path, drop_nas: bool = False):
    logger.debug("Gathering IDs from %s.", file_path)
    df = pd.read_csv(filepath_or_buffer=file_path)

    if drop_nas:
        df = df.dropna(how="any", axis=0)

    all_ids = tuple(df["ID"].astype(str))

    return all_ids


def _split_ids(
    all_ids: Sequence[str], valid_or_test_size: float | int = 0.1
) -> Tuple[Sequence[str], Sequence[str]]:
    assert len(all_ids) > 0
    train_ids, test_or_valid_ids = eir.data_load.label_setup.split_ids(
        ids=all_ids, valid_size=valid_or_test_size
    )

    train_ids_set = set(train_ids)
    test_ids_set = set(test_or_valid_ids)
    assert len(train_ids_set.intersection(test_ids_set)) == 0

    return train_ids, test_or_valid_ids


def _split_deeplake_ds_into_train_and_test(
    train_ids: Sequence[str], test_ids: Sequence[str], source: Path, destination: Path
) -> None:
    train_ids_set = set(train_ids)
    test_ids_set = set(test_ids)
    assert len(train_ids_set.intersection(test_ids_set)) == 0

    train_path = destination / "train"
    test_path = destination / "test"
    if train_path.exists() and test_path.exists():
        return

    ensure_path_exists(path=train_path, is_folder=True)
    ensure_path_exists(path=test_path, is_folder=True)

    full_deeplake_ds = load_deeplake_dataset(data_source=str(source))

    ds_train = deeplake.empty(path=train_path)
    ds_test = deeplake.empty(path=test_path)

    for name, tensor in full_deeplake_ds.tensors.items():
        ds_train.create_tensor(
            name=name,
            htype=tensor.htype,
            dtype=tensor.dtype,
            sample_compression=tensor.meta.sample_compression,
        )
        ds_test.create_tensor(
            name=name,
            htype=tensor.htype,
            dtype=tensor.dtype,
            sample_compression=tensor.meta.sample_compression,
        )

    skipped_samples = 0
    with full_deeplake_ds, ds_train, ds_test:
        for idx, sample in enumerate(full_deeplake_ds):
            cur_id = sample["ID"].numpy().item()
            if cur_id in train_ids_set:
                parsed_sample = _deeplake_sample_to_dict(sample=sample)
                ds_train.append(parsed_sample)
            elif cur_id in test_ids_set:
                parsed_sample = _deeplake_sample_to_dict(sample=sample)
                ds_test.append(parsed_sample)
            else:
                skipped_samples += 1

            if idx % 10000 == 0:
                logger.info(
                    "Iterated over %d samples while splitting deeplake dataset into "
                    "train and test. Skipped %d samples.",
                    idx,
                    skipped_samples,
                )


def _deeplake_sample_to_dict(sample: deeplake.Dataset) -> Dict:
    """
    This is a workaround for the fact that deeplake does creates empty samples for
    *some* dtypes when we try to add them directly when iterating over samples.
    For example, numpy arrays are converted correctly, but strings result in empty
    for some reason (possibly a bug in deeplake).

    Note that this is possibly because deeplake .append is type to
    support Dict[str, Any], so maybe it is not a bug, but it just partially works
    when we feed it a sample.
    """

    parsed = {}

    for name, tensor in sample.tensors.items():
        if tensor.htype == "text":
            parsed[name] = tensor.numpy().item()
        else:
            parsed[name] = tensor.numpy()

    return parsed


def _split_arrays_into_train_and_test(
    train_ids: Sequence[str], test_ids: Sequence[str], source: Path, destination: Path
) -> None:
    train_ids_set = set(train_ids)
    test_ids_set = set(test_ids)
    assert len(train_ids_set.intersection(test_ids_set)) == 0

    train_path = destination / "train"
    test_path = destination / "test"
    if train_path.exists() and test_path.exists():
        return

    ensure_path_exists(path=train_path, is_folder=True)
    ensure_path_exists(path=test_path, is_folder=True)

    for array_file in source.iterdir():
        if array_file.stem in train_ids_set:
            copyfile(src=array_file, dst=train_path / array_file.name)
        elif array_file.stem in test_ids_set:
            copyfile(src=array_file, dst=test_path / array_file.name)


def _split_csv_into_train_and_test(
    train_ids: Sequence[str], test_ids: Sequence[str], source: Path, destination: Path
) -> None:
    ensure_path_exists(path=destination, is_folder=True)
    df_labels = pd.read_csv(filepath_or_buffer=source)
    df_labels["ID"] = df_labels["ID"].astype(str)

    train_ids_set = set(train_ids)
    test_ids_set = set(test_ids)

    df_train = df_labels[df_labels["ID"].isin(train_ids_set)].copy()
    df_test = df_labels[df_labels["ID"].isin(test_ids_set)].copy()

    df_train.to_csv(destination / "labels_train.csv", index=False)
    df_test.to_csv(destination / "labels_test.csv", index=False)


class ParseDataWrapper(luigi.Task):
    data_config = luigi.DictParameter()

    def requires(self):
        yield CommonSplitIntoTestSet(**self.data_config)

    def output(self):
        return self.input()


def get_dynamic_valid_size(
    num_samples_per_epoch: int,
    minimum: int,
    valid_size_upper_bound: int = 20000,
) -> int:
    valid_size = int(0.1 * num_samples_per_epoch)

    if valid_size < minimum:
        valid_size = minimum

    if valid_size > valid_size_upper_bound:
        valid_size = valid_size_upper_bound

    return valid_size


def get_batch_size(
    samples_per_epoch: int,
    upper_bound: int = 64,
    lower_bound: int = 4,
) -> int:
    batch_size = 2 ** int(math.log2(samples_per_epoch / 20))

    if batch_size > upper_bound:
        return upper_bound
    elif batch_size < lower_bound:
        return lower_bound

    logger.info("Batch size set to: %d", batch_size)

    if batch_size <= 8:
        logger.warning(
            "Computed batch size based on number of training"
            " samples per epoch (%d) "
            " is very small (%d). This may cause issues with training."
            " This is likely due to a small number of samples in the dataset."
            " Consider increasing the number of samples in the dataset if possible.",
            samples_per_epoch,
            batch_size,
        )

    return batch_size



import argparse
from dataclasses import dataclass
from pathlib import Path

from eir_auto_gp.post_analysis.common.data_preparation import (
    DataPaths,
    ExperimentInfo,
    SplitModelData,
    build_data_paths,
    extract_experiment_info_from_config,
    set_up_split_model_data,
)
from eir_auto_gp.post_analysis.run_complexity_analysis import run_complexity_analysis
from eir_auto_gp.post_analysis.run_effect_analysis import run_effect_analysis


@dataclass()
class PostAnalysisObject:
    data_paths: DataPaths
    experiment_info: ExperimentInfo
    modelling_data: SplitModelData


def build_post_analysis_object(cl_args: argparse.Namespace) -> PostAnalysisObject:
    run_dir = Path(cl_args.run_dir)
    data_paths = build_data_paths(run_dir=run_dir)

    experiment_info = extract_experiment_info_from_config(
        config_path=data_paths.experiment_config
    )

    modelling_data = set_up_split_model_data(
        data_paths=data_paths,
        experiment_info=experiment_info,
        top_snps=cl_args.top_n_snps,
    )

    complexity_object = PostAnalysisObject(
        data_paths=data_paths,
        experiment_info=experiment_info,
        modelling_data=modelling_data,
    )

    return complexity_object


def get_cl_args() -> argparse.Namespace:
    parser = get_argument_parser()
    args = parser.parse_args()
    return args


def get_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--run_dir",
        type=str,
        help="Path to the run directory.",
    )

    parser.add_argument(
        "--top_n_snps",
        type=int,
        default=128,
        help="Number of SNPs to use for the analysis.",
    )

    return parser


def run_all():
    cl_args = get_cl_args()
    post_analysis_object = build_post_analysis_object(cl_args=cl_args)
    run_complexity_analysis(post_analysis_object=post_analysis_object)
    run_effect_analysis(post_analysis_object=post_analysis_object)


def main():
    run_all()


if __name__ == "__main__":
    main()



import warnings
from dataclasses import dataclass
from itertools import product
from typing import TYPE_CHECKING, Any, Dict, Generator

import numpy as np
import pandas as pd
import xgboost
from eir.train_utils import metrics as eir_metrics
from sklearn import metrics
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import ElasticNetCV, LogisticRegressionCV
from sklearn.preprocessing import LabelEncoder

if TYPE_CHECKING:
    from eir_auto_gp.post_analysis.run_complexity_analysis import ModelReadyObject

warnings.filterwarnings("ignore", category=ConvergenceWarning)


@dataclass()
class TrainEvalResults:
    df_predictions: pd.DataFrame
    df_predictions_raw: pd.DataFrame
    performance: pd.DataFrame


def train_and_evaluate_xboost(
    modelling_data: "ModelReadyObject",
    target_type: str,
) -> TrainEvalResults:
    df_target = modelling_data.target_train
    num_classes = df_target[df_target.columns[0]].nunique()
    params = get_xgboost_params(task_type=target_type, num_classes=num_classes)

    x_train = modelling_data.input_train.values
    y_train = modelling_data.target_train.values
    x_val = modelling_data.input_val.values
    y_val = modelling_data.target_val.values

    d_train = xgboost.DMatrix(data=x_train, label=y_train)
    d_val = xgboost.DMatrix(data=x_val, label=y_val)

    watchlist = [(d_train, "train"), (d_val, "eval")]
    trained_model = xgboost.train(
        params=params,
        dtrain=d_train,
        num_boost_round=10000,
        evals=watchlist,
        verbose_eval=100,
        early_stopping_rounds=100,
    )

    x_test = modelling_data.input_test.values
    y_test = modelling_data.target_test.values

    d_test = xgboost.DMatrix(data=x_test, label=y_test)
    y_score = trained_model.predict(data=d_test, output_margin=True)
    if target_type == "regression":
        y_score = y_score.reshape(-1, 1)

    performance = evaluate_model_performance(
        y_true=y_test,
        y_score=y_score,
        task_type=target_type,
    )
    df_performance = pd.DataFrame.from_dict(data=performance, orient="index").T

    df_predictions, df_predictions_raw = parse_predictions(
        modelling_data=modelling_data,
        y_test=y_test,
        y_score=y_score,
    )

    results = TrainEvalResults(
        df_predictions=df_predictions,
        df_predictions_raw=df_predictions_raw,
        performance=df_performance,
    )

    return results


def train_and_evaluate_linear(
    modelling_data: "ModelReadyObject",
    target_type: str,
) -> TrainEvalResults:
    if target_type not in ["classification", "regression"]:
        raise ValueError("target_type must be 'classification' or 'regression'")

    x_train = pd.concat([modelling_data.input_train, modelling_data.input_val]).values
    y_train = pd.concat([modelling_data.target_train, modelling_data.target_val]).values

    if target_type == "classification":
        model = LogisticRegressionCV(
            cv=5,
            random_state=0,
            max_iter=1000,
        )
    else:
        model = ElasticNetCV(
            eps=1e-7,
            cv=5,
            random_state=0,
            l1_ratio=[0.1, 0.5, 0.7, 0.9, 0.95, 0.99],
            tol=1e-03,
            selection="random",
        )

    model.fit(X=x_train, y=y_train.ravel())

    x_test = modelling_data.input_test.values
    y_test = modelling_data.target_test.values

    if target_type == "classification":
        y_score = model.predict_proba(x_test)
    else:
        y_score = model.predict(x_test)
        y_score = y_score.reshape(-1, 1)

    performance = evaluate_model_performance(
        y_true=y_test,
        y_score=y_score,
        task_type=target_type,
    )

    df_predictions, df_predictions_raw = parse_predictions(
        modelling_data=modelling_data,
        y_test=y_test,
        y_score=y_score,
    )
    df_performance = pd.DataFrame.from_dict(data=performance, orient="index").T

    results = TrainEvalResults(
        df_predictions=df_predictions,
        df_predictions_raw=df_predictions_raw,
        performance=df_performance,
    )

    return results


def parse_predictions(
    modelling_data: "ModelReadyObject",
    y_test: np.ndarray,
    y_score: np.ndarray,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    target_encoder = modelling_data.transformers.target_label_encoder
    if isinstance(target_encoder, LabelEncoder):
        class_names = target_encoder.classes_
        column_names = [f"Prediction_{class_name}" for class_name in class_names]
    else:
        column_names = ["Prediction"]

    df_predictions = pd.DataFrame(
        data=np.column_stack((y_test, y_score)),
        columns=["True Label"] + column_names,
        index=modelling_data.target_test.index,
    )

    if isinstance(target_encoder, LabelEncoder):
        y_score_raw = np.argmax(y_score, axis=1)
        y_test_raw = target_encoder.inverse_transform(y_test.ravel())
        y_score_raw = target_encoder.inverse_transform(y_score_raw.ravel())
    else:
        y_score_raw = y_score
        y_test_raw = y_test

    df_predictions_raw = pd.DataFrame(
        data=np.column_stack((y_test_raw, y_score_raw)),
        columns=["True Label", "Prediction"],
        index=modelling_data.target_test.index,
    )

    return df_predictions, df_predictions_raw


def evaluate_model_performance(
    y_true: np.ndarray, y_score: np.ndarray, task_type: str
) -> dict[str, float]:
    performance: dict[str, float]
    if task_type == "classification":
        acc = eir_metrics.calc_acc(outputs=y_score, labels=y_true)
        mcc = eir_metrics.calc_mcc(outputs=y_score, labels=y_true)
        roc_auc = eir_metrics.calc_roc_auc_ovo(outputs=y_score, labels=y_true)
        ap = eir_metrics.calc_average_precision(outputs=y_score, labels=y_true)
        performance = {
            "acc": acc,
            "mcc": mcc,
            "roc_auc": roc_auc,
            "ap": ap,
        }
    elif task_type == "regression":
        pcc = eir_metrics.calc_pcc(outputs=y_score, labels=y_true)
        r2 = eir_metrics.calc_r2(outputs=y_score, labels=y_true)
        rmse = metrics.mean_squared_error(y_true=y_true, y_pred=y_score, squared=False)
        performance = {
            "pcc": pcc,
            "r2": r2,
            "rmse": rmse,
        }
    else:
        raise ValueError()

    return performance


def get_xgboost_params(task_type: str, num_classes: int) -> Dict[str, Any]:
    assert task_type in {"classification", "regression"}

    if task_type == "classification":
        params = {
            "eta": 0.02,
            "max_depth": 7,
            "objective": "multi:softprob",
            "subsample": 0.5,
            "booster": "gbtree",
            "num_class": num_classes,
        }
    elif task_type == "regression":
        params = {
            "eta": 0.02,
            "max_depth": 7,
            "objective": "reg:squarederror",
            "subsample": 0.5,
            "booster": "gbtree",
        }
    else:
        raise ValueError()

    return params


def get_training_eval_iterator(
    any_tabular_input: bool,
) -> Generator[Dict[str, Any], None, None]:
    condition_lists = [
        ["xgboost", "linear"],
        [True, False],
        [True, False],
        [True, False],
    ]

    iterator = product(*condition_lists)
    for model_type, include_genotype, include_tabular, one_hot_encode in iterator:
        if not include_genotype and not include_tabular:
            continue
        if not include_genotype and one_hot_encode:
            continue
        if include_tabular and not any_tabular_input:
            continue

        yield {
            "model_type": model_type,
            "include_genotype": include_genotype,
            "include_tabular": include_tabular,
            "one_hot_encode": one_hot_encode,
        }
